/*
 * PUBLIC DOMAIN PCCTS-BASED C++ GRAMMAR (cplusplus.g, stat.g, expr.g)
 *
 * Authors: Sumana Srinivasan, NeXT Inc.;            sumana_srinivasan@next.com
 *          Terence Parr, Parr Research Corporation; parrt@parr-research.com
 *          Russell Quong, Purdue University;        quong@ecn.purdue.edu
 *
 * VERSION 1.2
 *
 * SOFTWARE RIGHTS
 *
 * This file is a part of the ANTLR-based C++ grammar and is free
 * software.  We do not reserve any LEGAL rights to its use or
 * distribution, but you may NOT claim ownership or authorship of this
 * grammar or support code.  An individual or company may otherwise do
 * whatever they wish with the grammar distributed herewith including the
 * incorporation of the grammar or the output generated by ANTLR into
 * commerical software.  You may redistribute in source or binary form
 * without payment of royalties to us as long as this header remains
 * in all source distributions.
 *
 * We encourage users to develop parsers/tools using this grammar.
 * In return, we ask that credit is given to us for developing this
 * grammar.  By "credit", we mean that if you incorporate our grammar or
 * the generated code into one of your programs (commercial product,
 * research project, or otherwise) that you acknowledge this fact in the
 * documentation, research report, etc....  In addition, you should say nice
 * things about us at every opportunity.
 *
 * As long as these guidelines are kept, we expect to continue enhancing
 * this grammar.  Feel free to send us enhancements, fixes, bug reports,
 * suggestions, or general words of encouragement at parrt@parr-research.com.
 * 
 * NeXT Computer Inc.
 * 900 Chesapeake Dr.
 * Redwood City, CA 94555
 * 12/02/1994
 * 
 * Restructured for public consumption by Terence Parr late February, 1995.
 *
 * Requires PCCTS 1.32b7 or higher to get past ANTLR. 
 * 
 * DISCLAIMER: we make no guarantees that this grammar works, makes sense,
 *             or can be used to do anything useful.
 */

#header <<
#include "MyZZPred.h"
#include "CPPDictionary.h"
#include "AToken.h"
typedef ANTLRCommonToken ANTLRToken;
>>

/* These token names are unnecessary in that the regular expressions
 * could easily have been placed into the grammar rules themselves.
 * However, if SORCERER is to be used to do any sort of translation
 * labels must be assigned for each token type.  Hence, I've left
 * the token label definitions here.
 */
#token LCURLY "\{"
#token RCURLY "\}"
#token LSQUARE "\["
#token RSQUARE "\]"
#token LPAREN "\("
#token RPAREN "\)"
#token COLON ":"
#token SEMICOLON ";"
#token COMMA ","
#token QUESTIONMARK "?"
#token ELLIPSIS "..."
#token ASSIGNEQUAL "="
#token TIMESEQUAL "\*="
#token DIVIDEEQUAL "/="
#token MODEQUAL "%="
#token PLUSEQUAL "\+="
#token MINUSEQUAL "\-="
#token SHIFTLEFTEQUAL "\<\<="
#token SHIFTRIGHTEQUAL "\>\>="
#token BITWISEANDEQUAL "&="
#token BITWISEXOREQUAL "^="
#token BITWISEOREQUAL "\|="
#token OR "\|\|"
#token AND "&&"
#token BITWISEOR "\|"
#token BITWISEXOR "^"
#token AMPERSAND "&"
#token EQUAL "=="
#token NOTEQUAL "!="
#token LESSTHAN "<"
#token GREATERTHAN ">"
#token LESSTHANOREQUALTO "<="
#token GREATERTHANOREQUALTO ">="
#token SHIFTLEFT "\<\<"
#token SHIFTRIGHT "\>\>"
#token PLUS "\+"
#token MINUS "\-"
#token STAR "\*"
#token DIVIDE "/"
#token MOD "%"
#token PLUSPLUS "\+\+"
#token MINUSMINUS "\-\-"
#token TILDE "\~"
#token NOT "!"
#token DOT "."
#token POINTERTO "\->"
#token POINTERTOMBR "\->\*"
#token DOTMBR ".\*"
#token SCOPE "::"

#token ASM "asm"
#token AUTO "auto"
#token BOOL "bool"
#token BREAK "break"
#token CASE "case"
#token CATCH "catch"
#token CHAR "char"
#token CLASS  "class"
#token CONST "const"
#token CONST "const_cast"
#token CONTINUE "continue"
#token DEFAULT "default"
#token DELETE "delete"
#token DO "do"
#token DOUBLE "double"
#token DYNAMIC_CAST "dynamic_cast"
#token ELSE "else"
#token ENUM "enum"
#token EXPLICIT "explicit"
#token EXTERN "extern"
#token BOOLFALSE "false"
#token FLOAT "float"
#token FOR "for"
#token FRIEND "friend"
#token GOTO "goto"
#token IF "if"
#token INLINE "inline"
#token INT "int"
#token LONG "long"
#token MUTABLE "mutable"
#token NAMESPACE "namespace"
#token NEW "new"
#token OPERATOR "operator"
#token PRIVATE "private"
#token PROTECTED "protected"
#token PUBLIC "public"
#token REGISTER "register"
#token REINTERPRET_CAST "reinterpret_cast"
#token RETURN "return"
#token SHORT "short"
#token SIGNED "signed"
#token SIZEOF "sizeof"
#token STATIC "static"
#token STATIC_CAST "static_cast"
#token STRUCT "struct"
#token SWITCH "switch"
#token TEMPLATE "template"
#token THIS "this"
#token THROW "throw"
#token BOOLTRUE "true"
#token TYPEDEF "typedef"
#token TYPEID "typeid"
#token TYPENAME "typename"
#token UNION "union"
#token UNSIGNED "unsigned"
#token USING "using"
#token VIRTUAL "virtual"
#token VOID "void"
#token VOLATILE "volatile"
#token WCHAR_T "wchar_t"
#token WHILE "while"

// Token class for operator types
#tokclass OPTOR_SIMPLE_TOKCLASS {
   PLUS MINUS STAR DIVIDE MOD BITWISEXOR AMPERSAND BITWISEOR TILDE NOT 
   SHIFTLEFT SHIFTRIGHT
   ASSIGNEQUAL TIMESEQUAL DIVIDEEQUAL MODEQUAL PLUSEQUAL MINUSEQUAL 
   SHIFTLEFTEQUAL SHIFTRIGHTEQUAL BITWISEANDEQUAL BITWISEXOREQUAL BITWISEOREQUAL
   EQUAL NOTEQUAL LESSTHAN GREATERTHAN LESSTHANOREQUALTO GREATERTHANOREQUALTO OR AND
   PLUSPLUS MINUSMINUS COMMA POINTERTO POINTERTOMBR
}


#token ID "[a-zA-Z_][a-zA-Z0-9_]*"
#token OCTALINT "0[0-7]*{[uUlL]}"
#token DECIMALINT "[1-9][0-9]*{[uUlL]}"
#token HEXADECIMALINT "(0x|0X)[0-9a-fA-F]+{[uUlL]}"
#token "'" << mode (CHARACTERS); more (); >>
#token "\"" << mode (STRINGS); more (); >>
#token FLOATONE "([0-9]+.[0-9]* | [0-9]*.[0-9]+) {[eE]{[\-\+]}[0-9]+} {[fFlL]}"
#token FLOATTWO "[0-9]+ [eE]{[\-\+]}[0-9]+ {[fFlL]}"

#token "/\*" << mode (COMMENT); skip (); >>

#token "[\t\ ]+"                 << skip (); >>
#token "[\n\r]"                  << newline(); skip(); >>
#token "// ~[\n]* \n"            << newline(); skip(); >>
#token "#pragma ~[\n]* \n"       << newline(); skip(); >>
// line number and file stuff from preprocessor
#token "#[\ \t]* [0-9]+ {[\ \t]* \"~[\"]+\" [\ \t]* [0-9]* [\ \t]* [0-9]*} \n"
                           << _line = atoi(begexpr()+1); skip(); >>
#token Eof "@"

#lexclass COMMENT
#token "[\n\r]"      << skip(); newline(); >>
#token "\*/"         << mode (START); skip (); >>
#token "\*"          << skip (); >>
#token "~[\*\n\r]+"  << skip (); >>

#lexclass STRINGS
#token STRING "\"" << mode (START); >>
#token "\\n" << replchar ((char) 0x0A); more (); >>
#token "\\t" << replchar ((char) 0x09); more (); >>
#token "\\v" << replchar ((char) 0x0B); more (); >>
#token "\\b" << replchar ((char) 0x08); more (); >>
#token "\\r" << replchar ((char) 0x0D); more (); >>
#token "\\f" << replchar ((char) 0x0C); more (); >>
#token "\\a" << replchar ((char) 0x07); more (); >>
#token "\\\\" << replchar ((char) 0x5C); more (); >>
#token "\\?" << replchar ((char) 0x3F); more (); >>
#token "\\'" << replchar ((char) 0x27); more (); >>
#token "\\\"" << replchar ((char) 0x22); more (); >>
#token "\\0[0-7]*" << replchar ((char) strtol (begexpr(), NULL, 8)); more (); >>
#token "\\[1-9][0-9]*" << replchar ((char) strtol (begexpr(), NULL, 10)); more (); >>
#token "\\(0x|0X)[0-9a-fA-F]+" << replchar ((char) strtol (begexpr(), NULL, 16)); more (); >>
#token "[\n\r]" << newline(); more (); >>
#token "~[\"\n\r\\]+" << more (); >>

#lexclass CHARACTERS 
#token CHARACTER "'" << mode (START); >>
#token "\\n" << replchar ((char) 0x0A); more (); mode (DONE); >>
#token "\\t" << replchar ((char) 0x09); more (); mode (DONE); >>
#token "\\v" << replchar ((char) 0x0B); more (); mode (DONE); >>
#token "\\b" << replchar ((char) 0x08); more (); mode (DONE); >>
#token "\\r" << replchar ((char) 0x0D); more (); mode (DONE); >>
#token "\\f" << replchar ((char) 0x0C); more (); mode (DONE); >>
#token "\\a" << replchar ((char) 0x07); more (); mode (DONE); >>
#token "\\\\" << replchar ((char) 0x5C); more (); mode (DONE); >>
#token "\\?" << replchar ((char) 0x3F); more (); mode (DONE); >>
#token "\\'" << replchar ((char) 0x27); more (); mode (DONE); >>
#token "\\\"" << replchar ((char) 0x22); more (); mode (DONE); >>
#token "\\0[0-7]*" << replchar ((char) strtol (begexpr(), NULL, 8)); more (); mode (DONE); >>
#token "\\[1-9][0-9]*" << replchar ((char) strtol (begexpr(), NULL, 10)); more (); mode (DONE); >>
#token "\\(0x|0X)[0-9a-fA-F]+" << replchar ((char) strtol (begexpr(), NULL, 16)); more (); mode (DONE); >>
#token "[\n\r]" << newline(); more (); >>
#token "~['\n\r\\]" << more (); mode (DONE); >>

#lexclass DONE
#token CHARACTER "'" << mode (START); >>


#lexclass START

<<
//typedef ANTLRCommonToken ANTLRToken;
>>

class CPPParser {
<<
public:
   #define CPPParser_MaxQualifiedItemSize 500

   // can't bitwise-OR enum elements together, this must be an int; damn!
   typedef unsigned long TypeSpecifier;   // note: must be > 16bits
   #define  tsInvalid  0x0
   #define tsVOID      0x1
   #define tsCHAR      0x2
   #define tsSHORT     0x4
   #define tsINT       0x8
   #define tsLONG      0x10
   #define tsFLOAT     0x20
   #define tsDOUBLE    0x40
   #define tsSIGNED    0x80
   #define tsUNSIGNED  0x100
   #define tsTYPEID    0x200
   #define tsSTRUCT    0x400
   #define tsENUM      0x800
   #define tsUNION     0x1000
   #define tsCLASS     0x2000
   #define tsWCHAR_T   0x4000
   #define tsBOOL      0x8000

   enum TypeQualifier { tqInvalid=0, tqCONST=1, tqVOLATILE };

   enum StorageClass {
      scInvalid=0, scAUTO=1, scREGISTER,
      scSTATIC, scEXTERN, scTYPEDEF
   };

   enum DeclSpecifier {
      dsInvalid=0,
      dsVIRTUAL, dsINLINE, dsFRIEND
   };

   // JEL 3/26/96 changed to allow ORing of values
   typedef int QualifiedItem;
   #define qiInvalid     0x0
   #define qiType        0x1         // includes enum, class, typedefs
   #define qiDtor        0x2
   #define qiCtor        0x4
   #define qiOperator    0x8
   #define qiPtrMember   0x10
   #define qiID          0x20        // not a type, but could be a var, func...

/*
   enum QualifiedItem {
      qiInvalid=0,
      qiType=1,         // includes enum, class, typedefs
      qiDtor, qiCtor, qiOperator, qiPtrMember,
      qiID           // not a type, but could be a var, func...
   };
*/

protected:
   // Symbol table management stuff
   CPPDictionary *symbols;
   StorageClass _sc;
   TypeQualifier _tq;
   TypeSpecifier _ts;
   unsigned char functionDefinition;
   int traceIndentLevel, doTracing;
   char qualifierPrefix[CPPParser_MaxQualifiedItemSize+1];
   int externalScope;   // global variables
   int templateParameterScope;
   char *enclosingClass;

   void tracein(char *r);
   void traceout(char *r);
   QualifiedItem qualifiedItemIs(int lookahead_offset=0);
   int qualifiedItemIsOneOf(QualifiedItem qiFlags, int lookahead_offset=0);
   char *qualified(char *);
   int scopedItem(int k=1);
   int finalQualifier(int k=1);
   int skipTemplateQualifiers(int& kInOut);
   int skipNestedParens(int& kInOut);

   // Limit lookahead for qualifiedItemIs()
   enum { MaxTemplateTokenScan = 200 };

public:
   void init()
   {
      ANTLRParser::init();
      symbols = new CPPDictionary(4001, 200, 800000);
      traceIndentLevel = 0;
      doTracing = 0;
      qualifierPrefix[0] = '\0';
      templateParameterScope = symbols->getCurrentScopeIndex();
      symbols->saveScope();
      externalScope = symbols->getCurrentScopeIndex();
      enclosingClass = "";
   }
   void traceOn() { doTracing=1; }

protected:

   // Semantic interface; subclass and redefine these functions
   // so you don't have to mess with the grammar itself.

   virtual int isTypeName(char *s);
   virtual int isClassName(char *s);
   virtual void enterNewLocalScope();
   virtual void exitLocalScope();
   virtual void enterExternalScope();
   virtual void exitExternalScope();

   // Aggregate stuff
   virtual void beginClassDefinition(TypeSpecifier, char *);
   virtual void endClassDefinition();
   virtual void classForwardDeclaration(TypeSpecifier, DeclSpecifier, char *);

   // Declaration stuff
   virtual void beginDeclaration();
   virtual void endDeclaration();
   virtual void beginFunctionDefinition();
   virtual void endFunctionDefinition();
   virtual void functionParameterList();
   virtual void functionEndParameterList();
   virtual void beginConstructorDefinition();
   virtual void endConstructorDefinition();
   virtual void beginConstructorDeclaration(char *);
   virtual void endConstructorDeclaration();
   virtual void beginDestructorDefinition();
   virtual void endDestructorDefinition();
   virtual void beginDestructorDeclaration(char *);
   virtual void endDestructorDeclaration();
   virtual void beginParameterDeclaration();
   virtual void beginFieldDeclaration();
   virtual void declarationSpecifier(StorageClass, TypeQualifier, TypeSpecifier);
   virtual void beginEnumDefinition(char *);
   virtual void endEnumDefinition();
   virtual void enumElement(char *);

   // Declarator stuff
   virtual void declaratorPointerTo();
   virtual void declaratorID(char *);
   virtual void declaratorArray();
   virtual void declaratorParameterList(int def);
   virtual void declaratorEndParameterList(int def);

   // template stuff
   virtual void templateTypeParameter(char *);
   virtual void beginTemplateDefinition();
   virtual void endTemplateDefinition();
   virtual void beginTemplateParameterList();
   virtual void endTemplateParameterList();

   // exception stuff
   virtual void exceptionBeginHandler();
   virtual void exceptionEndHandler();

   virtual void panic(char *);
>>

translation_unit
   :  <<enterExternalScope();>>
      (external_declaration)+ Eof
      <<exitExternalScope();>>
   ;

external_declaration
   :  
   // JEL 4/3/96 avoid hoisting predicates from alternates
   <<;>> <<;>>
   (

      // Class definition (templates too)
      // This is separated out otherwise the next alternative
      // would look for "class A { ... } f() {...}" which is
      // an unacceptable level of backtracking.

      // JEL Note:  Rule body does not need typedef, because
      // that is internal to "declaration", and it is invalid
      // to say "typedef template..."
      ( {TYPEDEF | template_head } class_head LCURLY )?
      { template_head } declaration
   |  
      // Enum definition (don't want to backtrack over this in other alts)
      ( ENUM {ID} LCURLY )?
      enum_specifier { init_declarator_list } SEMICOLON
   |  
      // Destructor DEFINITION (templated or non-templated)
      ( {template_head} dtor_decl_spec dtor_declarator[1] LCURLY )?
      {template_head} dtor_head dtor_body
   |  
      // Constructor DEFINITION (non-templated)
      //JEL 3/29/96 Does not account for leading inline.
      //(ID)? => <<qualifiedItemIsOneOf(qiCtor)>>?

      // JEL 4/3/96 Added predicate that works, once the
      // restriction is added that ctor cannot be virtual
      // and ctor_declarator uses a more restrictive id
      (SCOPE|ID)? => <<qualifiedItemIsOneOf(qiCtor)>>?
      (INLINE)? => <<qualifiedItemIsOneOf(qiCtor,1)>>?

      // JEL 4/3/96 Don't need to backtrack now that we have a good predicate
      //      ( 
      //         dtor_ctor_decl_spec
      //         ctor_declarator[1]
      //         { ctor_initializer } LCURLY
      //      )?
      ctor_definition 
   |  
      // User-defined type cast
      ( scope_override conversion_function_decl_or_def )?
      
   |
      // Function definition
      // Do not further hoist predicate from type_specifier
      <<;>> <<;>>
      (
		   { declaration_specifiers } 
         function_declarator LCURLY 
	   )?
      function_definition

   |  
      // templated forward class decl, init/decl of static member in template
      <<beginTemplateDefinition();>>
      (template_head declaration_specifiers { init_declarator_list } SEMICOLON)?
      <<endTemplateDefinition();>>

   |  
      // Templated FUNCTIONS and CONSTRUCTORS matched here.
      <<beginTemplateDefinition();>>
      template_head
      (  
            // Templated CONSTRUCTOR definition

            // JEL 4/3/96 Added predicate that works once the
            // restriction is added that ctor cannot be virtual
            (SCOPE|ID)? => <<qualifiedItemIsOneOf(qiCtor)>>?
            (INLINE)? => <<qualifiedItemIsOneOf(qiCtor,1)>>?

            // JEL 4/3/96 -- don't need to backtrack
            //            (
            //
            //               dtor_ctor_decl_spec 
            //               ctor_declarator[1]
            //               { ctor_initializer } LCURLY
            //            )?
            ctor_definition
         |  
            // Templated function definition
            { declaration_specifiers }
            function_declarator
            (
               compound_statement
               |   SEMICOLON
            )
      )
      <<endTemplateDefinition();>>

/*
      // stupid, but must allow "a;"
   |  init_declarator_list SEMICOLON
*/

      // everything else (except templates)
   |  declaration

      // believe it or not, this is a valid decl
   |  SEMICOLON

   )     // end of external_declaration

   ;

function_definition
   :  // don't want next action as an init-action due to (...)? caller
      <<beginFunctionDefinition();>>
      (  
            declaration_specifiers function_declarator compound_statement
         |  function_declarator compound_statement
      )
      <<endFunctionDefinition();>>
   ;

linkage_specification
   :  EXTERN
      STRING
      (  LCURLY (external_declaration)* RCURLY #pragma approx {SEMICOLON}
      |  declaration
      )
   ;

declaration
   :  <<;>>
      <<beginDeclaration();>>
      declaration_specifiers { init_declarator_list } SEMICOLON
      <<endDeclaration();>>
   |  linkage_specification
   ;

declaration_specifiers
   :  <<
      TypeSpecifier ts = tsInvalid, ts2;
      TypeQualifier tq = tqInvalid;
      StorageClass  sc = scInvalid;
      DeclSpecifier ds = dsInvalid;
      >>
      #pragma approx
      (  storage_class_specifier>[sc]
      |  type_specifier[ds]>[ts2]         <<ts |= ts2;>>
      |  type_qualifier>[tq]
      |  INLINE                <<ds = dsINLINE;>>
      |  VIRTUAL                  <<ds = dsVIRTUAL;>>
      |  FRIEND                  <<ds = dsFRIEND;>>
      )+
      <<declarationSpecifier(sc, tq, ts);>>
   ;

type_specifier[DeclSpecifier ds] > [CPPParser::TypeSpecifier ts]
   :  
      // Predicate to resolve ambiguities
      // involving declaration_specifiers.
      // We want predicate to be hoisted.
      (SCOPE | ID)? => <<qualifiedItemIsOneOf(qiType|qiCtor)>>?
      simple_type_specifier > [$ts]
   |  class_specifier[$ds] > [$ts]
   |  enum_specifier <<$ts=tsENUM;>>
   ;

simple_type_specifier > [CPPParser::TypeSpecifier ts]
   :  (  
            builtin_type_specifier > [$ts]
         |  qualified_type <<$ts=tsTYPEID;>>
      )
   ;

/* Match the A::B::C:: or nothing */
scope_override > [char *s]
   :  <<
      static char sitem[CPPParser_MaxQualifiedItemSize+1];
      sitem[0]='\0';
      >>
      {SCOPE <<strcat(sitem,"::");>> }

      #pragma approx
      (  
         <<scopedItem()>>?
         id:ID { LESSTHAN template_argument_list GREATERTHAN } SCOPE
         <<
         strcat(sitem,$id->getText());
         strcat(sitem,"::");
         >>
      )*
      <<$s = sitem;>>
   ;

// This matches a generic qualified identifier ::T::B::foo
// that is satisfactory for a ctor (no operator, no trailing <>)
qualified_ctor_id > [char *q]
   :  <<
      char *so;
      static char qitem[CPPParser_MaxQualifiedItemSize+1];
      >>
      scope_override > [so]
      <<strcpy(qitem, so);>>
      id2:ID
      <<strcat(qitem,$id2->getText());>>
      <<$q = qitem;>>
   ;

/* This matches a generic qualified identifier ::T::B::foo
 * (including "operator").
 * It might be a good idea to put T::~dtor in here
 * as well, but id_expression in expr.g puts it in manually.
 * Maybe not, 'cause many people use this assuming only A::B.
 * How about a 'qualified_complex_id'?
 */
qualified_id > [char *q]
   :  <<
      char *so;
      static char qitem[CPPParser_MaxQualifiedItemSize+1];
      >>
      scope_override > [so]
      <<strcpy(qitem, so);>>
      (  <<;>>
         id2:ID { LESSTHAN template_argument_list GREATERTHAN }
         <<strcat(qitem,$id2->getText());>>
      |  OPERATOR optor
         <<strcat(qitem,"operator"); strcat(qitem,"NYI");>>
      )
      <<$q = qitem;>>
   ;

/* Match A::B::* */
ptr_to_member :
   // JEL 4/4/96 Added SCOPE to context to allow ::A::*
   (SCOPE|ID)? => <<qualifiedItemIsOneOf(qiPtrMember)>>?
   scope_override STAR
   ;

qualified_type > [char *q]
   :  <<char *s; static char qitem[CPPParser_MaxQualifiedItemSize+1];>>

      // JEL 3/29/96 removed this predicate and moved it upwards to
	   // type_specifier.  This was done to allow parsing of ~ID to 
      // be a unary_expression, which was never reached with this 
      // predicate on
      // <<qualifiedItemIsOneOf(qiType|qiCtor)>>?
      
      scope_override > [s]
      id:ID 
      #pragma approx 
      { LESSTHAN template_argument_list GREATERTHAN }
      <<
      strcpy(qitem, s);
      strcat(qitem, $id->getText());
      $q = qitem;
      >>
   ;

type_qualifier > [CPPParser::TypeQualifier tq]
   :  CONST    <<$tq = tqCONST;>>
   |  VOLATILE <<$tq = tqVOLATILE;>>
   ;

storage_class_specifier > [CPPParser::StorageClass sc]
   :  AUTO     <<$sc = scAUTO;>>
   |  REGISTER <<$sc = scREGISTER;>>
   |  STATIC      <<$sc = scSTATIC;>>
   |  EXTERN      <<$sc = scEXTERN;>>
   |  TYPEDEF     <<$sc = scTYPEDEF;>>
   ;

builtin_type_specifier > [CPPParser::TypeSpecifier ts]
   :  VOID     <<$ts = tsVOID;>>
   |  CHAR     <<$ts = tsCHAR;>>
   |  WCHAR_T  <<$ts = tsWCHAR_T;>>
   |  SHORT    <<$ts = tsSHORT;>>
   |  INT      <<$ts = tsINT;>>
   |  LONG     <<$ts = tsLONG;>>
   |  FLOAT    <<$ts = tsFLOAT;>>
   |  DOUBLE   <<$ts = tsDOUBLE;>>
   |  SIGNED   <<$ts = tsSIGNED;>>
   |  UNSIGNED <<$ts = tsUNSIGNED;>>
   |  BOOL     <<$ts = tsBOOL;>>
   ;

typeID
   :  <<isTypeName(LT(1)->getText())>>?
      ID
   ;

init_declarator_list
   :  init_declarator (COMMA init_declarator)*
   ;

init_declarator
   :  declarator
      {  ASSIGNEQUAL initializer
      |  LPAREN expression_list RPAREN
      }
   ;

class_head
   :  (  STRUCT
      |  UNION
      |  CLASS
      )
      {ID { base_clause } }
   ;

class_specifier[DeclSpecifier ds] > [CPPParser::TypeSpecifier ts]
   :  <<char *saveClass;>>
      (  STRUCT   <<$ts=tsSTRUCT;>>
      |  UNION <<$ts=tsUNION;>>
      |  CLASS <<$ts=tsCLASS;>>
      )
      #pragma approx // indicate that ambig between alts 1 and 3 is ok
      (  id:ID
         <<saveClass = enclosingClass;
           enclosingClass = symbols->strdup($id->getText());
         >>
         { base_clause }
         LCURLY
            <<beginClassDefinition($ts, $id->getText());>>
            (member_declaration)*
            <<endClassDefinition();>>
         RCURLY
         <<enclosingClass = saveClass;>>
      |  LCURLY
            <<saveClass = enclosingClass; enclosingClass = "__anonymous";>>
            <<beginClassDefinition($ts, "anonymous");>>
            (member_declaration)*
            <<endClassDefinition();>>
         RCURLY
         <<enclosingClass = saveClass;>>

      |  tag:ID #pragma approx { LESSTHAN template_argument_list GREATERTHAN }
         <<classForwardDeclaration($ts, $ds, $tag->getText());>>
      )
   ;

base_clause
   :  COLON base_specifier (COMMA base_specifier)*
   ;

base_specifier
   :  (  VIRTUAL {access_specifier} ID
      |  access_specifier { VIRTUAL } ID
      |  ID
      )
      { LESSTHAN template_argument_list GREATERTHAN }
   ;

access_specifier
   :  PUBLIC
   |  PROTECTED
   |  PRIVATE
   ;

member_declaration
   :
   // JEL 4/3/96 -- avoid hoisting predicates
   <<;>> <<;>>
   (

      // Class definition
      // This is separated out otherwise the next alternative
      // would look for "class A { ... } f() {...}" which is
      // an unacceptable level of backtracking.
      ( {TYPEDEF} class_head LCURLY )? declaration
   |  
      // Enum definition (don't want to backtrack over this in other alts)
      ( ENUM {ID} LCURLY )?
      enum_specifier {member_declarator_list} SEMICOLON
      
   |  
      // JEL Predicate to distinguish ctor from function
      // This works now that ctor cannot have VIRTUAL
      // It unfortunately matches A::A where A is not enclosing
      // class -- this will have to be checked semantically.
      (SCOPE|ID)? => <<qualifiedItemIsOneOf(qiCtor)>>?
      (INLINE (SCOPE|ID))? => <<qualifiedItemIsOneOf(qiCtor,1)>>?
      (
         {INLINE} ctor_declarator[0]
         (
                ctor_body     // DEFINITION
            |   SEMICOLON           // DECLARATION
         )
      )
   |  
      // Destructor DECLARATION or DEFINITION
      // No template_head allowed for dtor member
      // Backtrack if not a dtor (no TILDE)
      (dtor_head)?
      dtor_head
      (
           dtor_body    // Definition
         | SEMICOLON          // Declaration
      )
   |  
      // Function definition
      <<beginFieldDeclaration();>>
      ( 
         { declaration_specifiers }
         function_declarator LCURLY 
      )?
      function_definition

   |  
      // User-defined type cast
      conversion_function_decl_or_def
      
   |  
      // Hack to handle decls like "superclass::member",
      // to redefine access to private base class public members
      ( qualified_id SEMICOLON )?

/*
      // stupid, but must allow member "a;"
      // JEL note:  This was probably commented out because it is
      // identical to the above rule.
   |  member_declarator_list SEMICOLON
*/

   |  
      // Member with a type or just a type def
      // A::T a(), ::T a, ::B a, void a, E a (where E is the enclosing class)
      (ID|SCOPE)? => <<qualifiedItemIsOneOf(qiType|qiCtor)>>?
      <<beginFieldDeclaration();>>
      declaration_specifiers {member_declarator_list} SEMICOLON

   |  
      // Member without a type (I guess it can only be a function decl)
      ( ID )? => <<1>>?
      <<beginFieldDeclaration();>>
      function_declarator SEMICOLON

   |  access_specifier COLON

   |  SEMICOLON
   
   )        // end member_declaration
   ;

member_declarator_list
   :  member_declarator { ASSIGNEQUAL OCTALINT } // The value must be 0 (pure virt.)
      ( COMMA member_declarator { ASSIGNEQUAL OCTALINT } )*
   ;

member_declarator
   :  declarator
   ;

conversion_function_decl_or_def
   :  OPERATOR declaration_specifiers { STAR | AMPERSAND }
      LPAREN { parameter_list } RPAREN
      {type_qualifier}
      (  compound_statement
      |  SEMICOLON
      )
   ;

enum_specifier
   :  ENUM
      (  LCURLY enumerator_list RCURLY
      |  id:ID
         <<beginEnumDefinition($id->getText());>>
         { LCURLY enumerator_list RCURLY }
         <<endEnumDefinition();>>
      )
   ;

enumerator_list
   :  enumerator (COMMA enumerator)*
   ;

enumerator
   :  id:ID { ASSIGNEQUAL constant_expression }
      <<enumElement($id->getText());>>
   ;

/* I think something is weird with the context-guards for predicates;
 * as a result I manually hoist the appropriate pred from ptr_to_member
 */
ptr_operator :
      AMPERSAND cv_qualifier_seq
   |  STAR cv_qualifier_seq
   |  (ID)? => <<qualifiedItemIsOneOf(qiPtrMember)>>?
      <<;>> ptr_to_member cv_qualifier_seq
   ;

// JEL note:  does not use (const|volatile)* to avoid lookahead problems
cv_qualifier_seq
   :  CONST
   |  CONST VOLATILE
   |  VOLATILE
   |  VOLATILE CONST
   |  /* empty */
   ;

declarator
   :  ptr_operator declarator <<declaratorPointerTo();>>
   |  ( ID )? => <<1>>? // counteract qiPtrMember pred in ptr_operator
      direct_declarator
   ;

direct_declarator
   :  <<char *id;>>
      qualified_id > [id] <<declaratorID(id);>>
      declarator_suffixes

   |  TILDE dtor:ID <<declaratorID($dtor->getText());>>
      declarator_suffixes

   |  LPAREN declarator RPAREN declarator_suffixes
   ;

declarator_suffixes
   :  ( LSQUARE { constant_expression } RSQUARE )+
      <<declaratorArray();>>
   |  (LPAREN ID)? => <<qualifiedItemIsOneOf(qiType|qiCtor,1)>>?
      LPAREN <<declaratorParameterList(0);>>
         { parameter_list }
      RPAREN <<declaratorEndParameterList(0);>>
      { type_qualifier }
   |
   ;

/* I think something is weird with the context-guards for predicates;
 * as a result I manually hoist the appropriate pred from ptr_to_member
 *
 * TER: warning: seems that "ID::" will always bypass and go to 2nd alt :(
 */
function_declarator
   :  (ID)? => <<qualifiedItemIsOneOf(qiPtrMember)>>?
      <<;>> ptr_operator function_declarator <<declaratorPointerTo();>>
   |  function_direct_declarator
   ;

function_direct_declarator
   :  <<char *q;>>
      /* predicate indicate that plain ID is ok here; this counteracts any
       * other predicate that gets hoisted (along with this one) that
       * indicates that an ID is a type or whatever.  E.g.,
       * another rule testing isTypeName() alone, implies that the
       * the ID *MUST* be a type name.  Combining isTypeName() and
       * this predicate in an OR situation like this one:
       * ( declaration_specifiers ... | function_declarator ... )
       * would imply that ID can be a type name OR a plain ID.
       */
      ( ID )? => <<1>>?
      qualified_id>[q]
      <<
      declaratorID(q);
      functionParameterList();
      >>
      LPAREN { parameter_list } RPAREN
      { type_qualifier }
      { ASSIGNEQUAL OCTALINT }     // The value of the octal must be 0
      <<functionEndParameterList();>>
   ;

// Used to be dtor_ctor_decl_spec
dtor_decl_spec
   :  INLINE
   |  VIRTUAL
   |  VIRTUAL INLINE
   |  INLINE VIRTUAL
   |
   ;

// JEL 4/3/96 does not include leading {template_head} 
// needed for external dtor definitions
dtor_head:
   dtor_decl_spec 
   dtor_declarator[1]
   ;

// JEL 4/3/96
dtor_body:
   compound_statement
   ;

ctor_head :
   {INLINE} ctor_declarator[1]
   ;

ctor_body :
   { ctor_initializer }
   compound_statement
   ;

ctor_definition :
   ctor_head
   ctor_body
   ;

// JEL 4/3/96replaced this
//ctor_definition :
//   :  <<;>>
//      <<beginConstructorDefinition(LT(1)->getText());>>
//
//      // JEL 4/3/96 ctor cannot be virtual.  This restriction
//      // reduces the complexity of ctor predicates
//      // dtor_decl_spec 
//      { INLINE }
//
//      ctor_declarator[1] 
//      { ctor_initializer }
//      compound_statement
//      <<endConstructorDefinition();>>
//   ;

dtor_declarator[int definition]  :  
   scope_override 
   TILDE ID
   <<declaratorParameterList($definition);>>
   // JEL 4/3/96 dtor has no args
   LPAREN RPAREN
   <<declaratorEndParameterList($definition);>>
   ;

ctor_declarator[int definition]
   :  
      // JEL 4/3/96 qualified_id too broad
      qualified_ctor_id
      <<declaratorParameterList($definition);>>
      LPAREN { parameter_list } RPAREN
      <<declaratorEndParameterList($definition);>>
   ;

ctor_initializer
   :  COLON superclass_init (COMMA superclass_init)*
   ;

superclass_init
   :  qualified_id LPAREN {expression_list} RPAREN
   ;

parameter_list
   :  parameter_declaration_list { {COMMA} ELLIPSIS }
   |  ELLIPSIS
   ;

parameter_declaration_list
   :  parameter_declaration ( COMMA parameter_declaration )*
   ;

parameter_declaration
   :  <<;>>
      <<beginParameterDeclaration();>>
      declaration_specifiers
      (  (declarator)?        // if arg name given
      |  abstract_declarator     // if arg name not given
      )
      { ASSIGNEQUAL assignment_expression}
   ;

initializer
   :  assignment_expression
   |  LCURLY initializer (COMMA initializer)* RCURLY
   ;

type_name
   :  declaration_specifiers abstract_declarator
   ;

/* This rule looks a bit weird because (...) can happen in two
 * places within the declaration such as "void (*)()" (ptr to
 * function returning nothing).  However, the () of a function
 * can only occur after having seen either a (abstract_declarator)
 * and not after a [..] or simple '*'.  These are the only two
 * valid () func-groups:
 *    int (*)();     // ptr to func
 *    int (*[])();   // array of ptr to func
 */
abstract_declarator
   :  ptr_operator abstract_declarator <<declaratorPointerTo();>> 
   |  LPAREN abstract_declarator RPAREN
      (abstract_declarator_suffix)+
   |  (  LSQUARE { constant_expression } RSQUARE
         <<declaratorArray();>>
      )+
   |  /* empty */
   ;

abstract_declarator_suffix
   :  LSQUARE { constant_expression } RSQUARE
      <<declaratorArray();>>
   |  LPAREN
         <<declaratorParameterList(0);>>
         { parameter_list }
      RPAREN
      <<declaratorEndParameterList(0);>>
   ;

template_head
   :  TEMPLATE
      <<beginTemplateParameterList();>>
      LESSTHAN template_parameter_list GREATERTHAN
      <<endTemplateParameterList();>>
   ;

/*
template_declaration
   :  <<;>>
      <<beginTemplateDefinition();>>
      template_head declaration
      <<endTemplateDefinition();>>
   ;
*/

template_parameter_list
   :  template_parameter ( COMMA template_parameter )*
   ;

/* Rule requires >2 lookahead tokens. The ambiguity is resolved 
 * correctly, however. According to the manual ''...A template argument
 * that can be interpreted either as a parameter-declaration or a
 * type-argument (because its identifier is the name of an
 * already exisitng class) is taken as type-argument.''
 * Therefore, any "class ID" that is seen on the input, should
 * match the first alternative here (it should be a type-argument).
 */
template_parameter
   :  #pragma approx // make ANTLR shut up about this ambiguity
      (  CLASS id:ID  <<templateTypeParameter($id->getText());>>
      |  parameter_declaration
      )
   ;

/* This rule refers to an instance of a template class or function */
template_id
   :  ID LESSTHAN template_argument_list GREATERTHAN
   ;

template_argument_list
   :  template_argument ( COMMA template_argument )*
   ;

/* Here assignment_expression was changed to shift_expression to rule out
 *  x< 1<2 > which causes ambiguities. As a result, these can be used only
 *  by enclosing parentheses x<(1<2)>. This is true for x<1+2> ==> bad,
 *  x<(1+2)>==>ok.
 */
template_argument
   :  type_name
   |  shift_expression
   ;




///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
//////////////////////////////  STATEMENTS ////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////


statement_list
	:	(statement)+
	;

statement
	:
	( declaration )?
   |  labeled_statement
	|	expression SEMICOLON
	|	compound_statement
	|	selection_statement
	|	iteration_statement
	|	jump_statement
	|	SEMICOLON
	|	try_block
	|	throw_statement
	;

labeled_statement
	:	ID COLON statement
	|	CASE constant_expression COLON statement
	|	DEFAULT COLON statement
	;

compound_statement
	:	LCURLY <<enterNewLocalScope();>>
		{ statement_list }
		RCURLY <<exitLocalScope();>>
	;

/* NOTE: cannot remove ELSE ambiguity, but it parses correctly.
 * The warning is removed with the #pragma
 */
selection_statement
	:	IF LPAREN expression RPAREN statement
		#pragma approx
		{ ELSE statement }
	|	SWITCH LPAREN expression RPAREN statement
	;

iteration_statement
	:	WHILE LPAREN expression RPAREN statement
	|	DO statement WHILE LPAREN expression RPAREN SEMICOLON
	|	FOR LPAREN
			( ( declaration )? | expression SEMICOLON | SEMICOLON )
			{ expression } SEMICOLON
			{ expression }
		RPAREN statement
	;

jump_statement
	:	GOTO ID SEMICOLON
	|	CONTINUE SEMICOLON
	|	BREAK SEMICOLON
	|  	RETURN { expression } SEMICOLON
	;

try_block
	:	TRY compound_statement (handler)*
	;

handler
	:	CATCH
		<<exceptionBeginHandler();>>
		<<declaratorParameterList(1);>>
		LPAREN exception_declaration RPAREN
		<<declaratorEndParameterList(1);>>
		compound_statement
		<<exceptionEndHandler();>>
	;

exception_declaration
	:	parameter_declaration_list
	|	ELLIPSIS
	;

/* This is an expression of type void according to the ARM, which
 * to me means "statement"; it removes some ambiguity to put it in
 * as a statement also.
 */
throw_statement
	:	THROW {assignment_expression} SEMICOLON
	;

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
//////////////////////////////  EXPRESSIONS ///////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

expression
   :  assignment_expression (COMMA assignment_expression)*
   ;

/* right-to-left for assignment op */
assignment_expression
   :  conditional_expression
      {  (  ASSIGNEQUAL
         |  TIMESEQUAL
         |  DIVIDEEQUAL
         |  MODEQUAL
         |  PLUSEQUAL
         |  MINUSEQUAL
         |  SHIFTLEFTEQUAL
         |  SHIFTRIGHTEQUAL
         |  BITWISEANDEQUAL
         |  BITWISEXOREQUAL
         |  BITWISEOREQUAL
         )
         assignment_expression
      }
   ;

conditional_expression
   :  logical_or_expression
      { QUESTIONMARK logical_or_expression COLON logical_or_expression }
   ;

constant_expression
   :  conditional_expression
   ;

logical_or_expression
   :  logical_and_expression (OR logical_and_expression)*
   ;

logical_and_expression
   :  inclusive_or_expression (AND inclusive_or_expression)*
   ;

inclusive_or_expression
   :  exclusive_or_expression (BITWISEOR exclusive_or_expression)*
   ;

exclusive_or_expression
   :  and_expression (BITWISEXOR and_expression)*
   ;

and_expression
   :  equality_expression (AMPERSAND equality_expression)*
   ;

equality_expression
   :  relational_expression ((NOTEQUAL | EQUAL) relational_expression)*
   ;

relational_expression
   :  shift_expression
      #pragma approx
      (  (  LESSTHAN
         |  GREATERTHAN
         |  LESSTHANOREQUALTO
         |  GREATERTHANOREQUALTO
         )
         shift_expression
      )*
   ;

shift_expression
   :  additive_expression ((SHIFTLEFT | SHIFTRIGHT) additive_expression)*
   ;

/* See comment for multiplicative_expression regarding #pragma */
additive_expression
   :  multiplicative_expression
      #pragma approx
      ((PLUS | MINUS) multiplicative_expression)*
   ;

/* ANTLR has trouble dealing with the analysis of the confusing unary/binary
 * operators such as STAR, AMPERSAND, PLUS, etc...  With the #pragma
 * we simply tell ANTLR to use the "quick-to-analyze" approximate lookahead
 * as full LL(k) lookahead will not resolve the ambiguity anyway.  Might
 * as well not bother.  This has the side-benefit that ANTLR doesn't go
 * off to lunch here (take infinite time to read grammar).
 */
multiplicative_expression
   :  pm_expression
      #pragma approx 
	  ((STAR | DIVIDE | MOD) pm_expression)*
   ;

pm_expression
   :  cast_expression (( DOTMBR | POINTERTOMBR ) cast_expression)*
   ;

/* The string "( ID" can be either the start of a cast or
 * the start of a unary_expression.  However, the ID must
 * be a type name for it to be a cast.  Since ANTLR can only hoist
 * semantic predicates that are visible without consuming a token,
 * the semantic predicate in rule type_name is not hoisted--hence, the
 * rule is reported to be ambiguous.  I am manually putting in the
 * correctly hoisted predicate.
 *
 * Ack! Actually "( ID" might be the start of "(T(expr))" which makes
 * the first parens just an ordinary expression grouping.  The solution
 * is to look at what follows the type, T.  Note, this could be a
 * qualified type.  Yucko.  I believe that "(T(" can only imply
 * function-style type cast in an expression (...) grouping.
 *
 * We DO NOT handle the following situation correctly at the moment:
 * Suppose you have
 *    struct rusage rusage;
 *    return (rusage.fp);
 *    return (rusage*)p;
 * Now essentially there is an ambiguity here. If rusage is followed by any
 * postix operators then it is an identifier else it is a type name. This
 * problem does not occur in C because, unless the tag struct is attached,
 * rusage is not a type name. However in C++ that restriction is removed.
 * No *real* programmer would do this, but it's in the C++ standard just for
 * fun..
 *
 * Another fun one (from an LL standpoint):
 *
 *   (A::B::T *)v;      // that's a cast of v to type A::B::T
 *   (A::B::foo);    // that's a simple member access
 *
 * The qualifiedItemIs(1) function scans ahead to what follows the
 * final "::" and returns qiType if the item is a type.  The offset of
 * '1' makes it ignore the initial LPAREN; normally, the offset is 0.
 */
cast_expression 
   :
   // JEL 4/2/96 Inhibit hoisting of predicate above this rule.
   // Otherwise, "(i<3);" does not qualify as a statement.
   <<;>> <<;>>
   (
         // JEL NOTE: The test for LT(3) in this predicate is not
         // completely correct, because it will miss (A::B(
         // instead, the predicate should check for "(" following the
         // (perhaps qualified) type name.
         ( LPAREN SCOPE | LPAREN ID )? =>
         <<qualifiedItemIsOneOf(qiType|qiCtor,1)&&LT(3)->getType()!=LPAREN>>?
         LPAREN type_name RPAREN cast_expression
      |  
         unary_expression  // handles outer (...) of "(T(expr))"
   )
   ;

unary_expression
   :
	// JEL 3/31/96 avoid hoisting predicate from postfix_expression -- 
   // it gets hauled all the way up to the "statement" rule and causes errors
   // on TILDE ID.
	<<;>> <<;>>		
					
	(  
         postfix_expression
	   |  PLUSPLUS unary_expression
	   |  MINUSMINUS unary_expression
	   |  unary_operator cast_expression
	   |  
		  SIZEOF
		  (  
			 // see comment for rule cast_expression for info on predicate
          // JEL NOTE 3/31/96 -- This won't work -- you really need to
          // call qualifiedItemIsOneOf(qiType|qiCtor,1)
          // The context should also be ( LPAREN (SCOPE|ID) )
				( LPAREN ID )? => <<isTypeName(LT(2)->getText())>>?
				LPAREN type_name RPAREN
			 |  unary_expression
		  )
	   |  new_expression
	   |  delete_expression
	)
   ;

/* The first ()? is used to resolve "new (expr) (type)" because both
 * (expr) and (type) look identical until you've seen the whole thing.
 *
 * new_initializer appears to be conflicting with function arguments as
 * function arguments can follow a primary_expression.  [This is a full
 * LL(k) versus LALL(k) problem.  Enhancing context by duplication of
 * some rules might handle this.]
 */
new_expression
   :  // counteract pred for postfix expr that says :: must start a
      // qualified item; here, it can start a "new".
      ( SCOPE )? => <<1>>?     

      { SCOPE } NEW
      (  (LPAREN expression_list RPAREN)?
      |
      )
      (  new_type_id
      |  LPAREN type_name RPAREN        /* new (T(3)); */
      )
      ( (new_initializer)? | )
   ;

new_type_id
   :  declaration_specifiers #pragma approx { new_declarator }
   ;

new_declarator
   :  ptr_to_member cv_qualifier_seq #pragma approx { new_declarator }
   |  direct_new_declarator
   ;

/* The "[expression]" construct conflicts with the "new []" construct
 * (and possibly others).  We used approximate lookahead for the "new []"
 * construct so that it would not try to compute full LL(2) lookahead.
 * Here, we use #pragma approx again because anytime we see a [ followed
 * by token that can begin an expression, we always want to loop.
 * Approximate lookahead handles this correctly.  In fact, approximate
 * lookahead is the same as full lookahead when all but the last lookahead
 * depth are singleton sets; e.g., {"["} followed by FIRST(expression).
 */
direct_new_declarator
   :  #pragma approx
      ( LSQUARE expression RSQUARE )+
   ;

new_initializer
   :  LPAREN { expression_list } RPAREN
   ;

delete_expression
   :  { SCOPE } DELETE { LSQUARE RSQUARE } cast_expression
   ;

unary_operator
   :  AMPERSAND
   |  STAR
   |  PLUS
   |  MINUS
   |  TILDE
   |  NOT
   ;

postfix_expression
   :  
      // JEL 4/1/96 We want to hoist this predicate up to unary_expression
      (TILDE ID)? => <<qualifiedItemIsOneOf(qiID|qiDtor|qiCtor)>>?
      
      // JEL 4/1/96 Don't hoist disambiguating predicate
      <<;>>

      #pragma approx
      (  
            // Function-style cast must have a leading type
            (ID|SCOPE)? => <<qualifiedItemIsOneOf(qiType|qiCtor)>>?
            simple_type_specifier LPAREN { expression_list } RPAREN
         |  
            primary_expression
            #pragma approx
            (  
                  LSQUARE expression RSQUARE
               |  LPAREN { expression_list } RPAREN
               |  DOT id_expression
               |  POINTERTO id_expression
               |  PLUSPLUS
               |  MINUSMINUS
            )*
      )
   ;

id_expression :
    scope_override
    (  
       ID
       |  OPERATOR optor
       | TILDE ID
    )
    ;

primary_expression
   :  id_expression
   |  constant
   |  THIS
   |  STRING
   |  LPAREN expression RPAREN
   ;

expression_list
   :  assignment_expression (COMMA assignment_expression)*
   ;

constant
   :  OCTALINT
   |  DECIMALINT
   |  HEXADECIMALINT
   |  CHARACTER
   |  FLOATONE
   |  FLOATTWO
   |  BOOLTRUE
   |  BOOLFALSE
   ;

optor :
      NEW
      #pragma approx 
      ( LSQUARE RSQUARE | )
   |   
      DELETE
      #pragma approx 
      ( LSQUARE RSQUARE | )
   |  LPAREN RPAREN
   |  LSQUARE RSQUARE
   |  
      declaration_specifiers  // user-defined type casts (ugh)
      #pragma approx
      { STAR | AMPERSAND }
   |  OPTOR_SIMPLE_TOKCLASS
   ;

}
