/*
 * PUBLIC DOMAIN PCCTS-BASED C++ GRAMMAR (cplusplus.g, stat.g, expr.g)
 *
 * Authors: Sumana Srinivasan, NeXT Inc.;            sumana_srinivasan@next.com
 *          Terence Parr, Parr Research Corporation; parrt@parr-research.com
 *          Russell Quong, Purdue University;        quong@ecn.purdue.edu
 *
 * VERSION 1.1
 *
 * SOFTWARE RIGHTS
 *
 * This file is a part of the ANTLR-based C++ grammar and is free
 * software.  We do not reserve any LEGAL rights to its use or
 * distribution, but you may NOT claim ownership or authorship of this
 * grammar or support code.  An individual or company may otherwise do
 * whatever they wish with the grammar distributed herewith including the
 * incorporation of the grammar or the output generated by ANTLR into
 * commerical software.  You may redistribute in source or binary form
 * without payment of royalties to us as long as this header remains
 * in all source distributions.
 *
 * We encourage users to develop parsers/tools using this grammar.
 * In return, we ask that credit is given to us for developing this
 * grammar.  By "credit", we mean that if you incorporate our grammar or
 * the generated code into one of your programs (commercial product,
 * research project, or otherwise) that you acknowledge this fact in the
 * documentation, research report, etc....  In addition, you should say nice
 * things about us at every opportunity.
 *
 * As long as these guidelines are kept, we expect to continue enhancing
 * this grammar.  Feel free to send us enhancements, fixes, bug reports,
 * suggestions, or general words of encouragement at parrt@parr-research.com.
 * 
 * NeXT Computer Inc.
 * 900 Chesapeake Dr.
 * Redwood City, CA 94555
 * 12/02/1994
 * 
 * Restructured for public consumption by Terence Parr late February, 1995.
 *
 * Requires PCCTS 1.32b4 or higher to get past ANTLR. 
 * 
 * DISCLAIMER: we make no guarantees that this grammar works, makes sense,
 *             or can be used to do anything useful.
 */

/*
 * Main program to test C++ grammar.
 */

#include <iostream.h>
#include "tokens.h"
#include "CPPParser.h"
typedef ANTLRCommonToken ANTLRToken;
#include "DLGLexer.h"
#include "PBlackBox.h"

/* I have subclassed the parser here so that the name of the
 * functions that are triggered are printed to stdout.
 */

class MyParser : public CPPParser {
public:
	MyParser(ANTLRTokenBuffer *input) : CPPParser(input) {;}

	virtual void enterNewLocalScope()
		{ fprintf(stdout, "enterNewLocalScope();\n"); CPPParser::enterNewLocalScope(); }
	virtual void exitLocalScope()
		{fprintf(stdout, "exitLocalScope();\n"); CPPParser::exitLocalScope();}

	// Declaration stuff
	virtual void beginDeclaration()
		{fprintf(stdout, "beginDeclaration();\n"); CPPParser::beginDeclaration();}
	virtual void beginConstructorDeclaration(char *ctor)
		{
			fprintf(stdout, "beginConstructorDeclaration(%s);\n", ctor);
			CPPParser::beginConstructorDeclaration(ctor);
		}
	virtual void beginDestructorDeclaration(char *dtor)
		{
			fprintf(stdout, "beginDestructorDeclaration(%s);\n", dtor);
			CPPParser::beginDestructorDeclaration(dtor);
		}
	virtual void beginFunctionDefinition()
		{
			fprintf(stdout, "beginFunctionDefinition();\n");
			CPPParser::beginFunctionDefinition();
		}
	virtual void beginConstructorDefinition()
		{
			fprintf(stdout, "beginConstructorDefinition\n");
			CPPParser::beginConstructorDefinition();
		}
	virtual void beginDestructorDefinition()
		{
			fprintf(stdout, "beginDestructorDefinition\n");
			CPPParser::beginDestructorDefinition();
		}
	virtual void beginEnumDefinition(char *e)
		{
			fprintf(stdout, "beginEnumDefinition(%s);\n", e);
			CPPParser::beginEnumDefinition(e);
		}
	virtual void endEnumDefinition()
		{
			fprintf(stdout, "endEnumDefinition();\n");
			CPPParser::endEnumDefinition();
		}
	virtual void beginParameterDeclaration()
		{
			fprintf(stdout, "beginParameterDeclaration();\n");
			CPPParser::beginParameterDeclaration();
		}
	virtual void beginFieldDeclaration()
		{
			fprintf(stdout, "beginFieldDeclaration();\n");
			CPPParser::beginFieldDeclaration();
		}
	virtual void declarationSpecifier(StorageClass sc,
									  TypeQualifier tq,
									  TypeSpecifier ts)
		{
			fprintf(stdout, "declarationSpecifier();\n");
			CPPParser::declarationSpecifier(sc,tq,ts);
		}

	// Declarator stuff
	virtual void declaratorPointerTo()
		{
			fprintf(stdout, "declaratorPointerTo();\n");
			CPPParser::declaratorPointerTo();
		}
	virtual void declaratorID(char *id)
		{
			fprintf(stdout, "declaratorID(%s);\n", id);
			CPPParser::declaratorID(id);
		}
	virtual void declaratorArray()
		{
			fprintf(stdout, "declaratorArray();\n");
			CPPParser::declaratorArray();
		}
	virtual void declaratorParameterList(int def)
		{
			fprintf(stdout, "declaratorParameterList();\n");
			CPPParser::declaratorParameterList(def);
		}
	virtual void declaratorEndParameterList(int def)
		{
			fprintf(stdout, "declaratorEndParameterList();\n");
			CPPParser::declaratorEndParameterList(def);
		}
	virtual void beginClassDefinition(TypeSpecifier ts, char *tag)
	{
		fprintf(stdout, "beginClassDefinition(%s)\n", tag);
		CPPParser::beginClassDefinition(ts,tag);
	}
	void templateTypeParameter(char *t)
		{
			fprintf(stdout, "templateTypeParameter(%s)\n", t);
			CPPParser::templateTypeParameter(t);
		}
	void beginTemplateDefinition()
		{
			fprintf(stdout, "beginTemplateDefinition()\n");
			CPPParser::beginTemplateDefinition();
		}
	void endTemplateDefinition()
		{
			fprintf(stdout, "endTemplateDefinition()\n");
			CPPParser::endTemplateDefinition();
		}
	void beginTemplateParameterList()
		{
			fprintf(stdout, "beginTemplateParameterList()\n");
			CPPParser::beginTemplateParameterList();
		}
	void endTemplateParameterList()
		{
			fprintf(stdout, "endTemplateParameterList()\n");
			CPPParser::endTemplateParameterList();
		}
	void exceptionBeginHandler()
		{
			fprintf(stdout, "exceptionBeginHandler()\n");
			CPPParser::exceptionBeginHandler();
		}
	void exceptionEndHandler()
		{
			fprintf(stdout, "exceptionEndHandler()\n");
			CPPParser::exceptionEndHandler();
		}
};

main(int argc, char *argv[])
{
   // ParserBlackBox<DLGLexer, CPPParser, ANTLRToken> p(stdin);
	ParserBlackBox<DLGLexer, MyParser, ANTLRToken> p(stdin);
	if ( argc>1 )
	{
		if ( strcmp("-trace", argv[1])==0 ) p.parser()->traceOn();
	}
	//	p.parser()->deleteTokens();
   p.parser()->translation_unit();
   return 0;
}
